/****************************************************************************
** iconset.cpp - various graphics handling classes
** Copyright (C) 2001, 2002  Justin Karneges
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
**
****************************************************************************/

#include"iconset.h"
#include"common.h"
#include"zip.h"
#include"psievent.h"

#include<qfile.h>
#include<qfileinfo.h>
#include<qtimer.h>

using namespace Jabber;

/****************************************************************************
  Impix - stores an image
****************************************************************************/
Impix::Impix()
{
	v_pixmap = 0;
	v_image = 0;
}

Impix::Impix(const QPixmap &x)
{
	v_pixmap = new QPixmap(x);
	v_image = new QImage(x.convertToImage());
}

Impix::Impix(const QImage &x)
{
	v_image = new QImage(x);
	v_pixmap = new QPixmap;
	v_pixmap->convertFromImage(*v_image);
}

Impix::~Impix()
{
	unload();
}

Impix::Impix(const Impix &from)
{
	v_pixmap = 0;
	v_image = 0;
	*this = from;
}

Impix & Impix::operator=(const Impix &from)
{
	if(from.v_pixmap)
		setPixmap(*from.v_pixmap);
	if(from.v_image)
		setImage(*from.v_image);
	return *this;
}

bool Impix::isNull() const
{
	if(!v_pixmap)
		return true;
	else
		return false;
}

const QPixmap & Impix::pixmap() const
{
	//if(v_pixmap)
		return *v_pixmap;
	//else
	//	return QPixmap();
}

const QImage & Impix::image() const
{
	//if(v_image)
		return *v_image;
	//else
	//	return QImage();
}

void Impix::setPixmap(const QPixmap &x)
{
	if(v_pixmap)
		delete v_pixmap;
	v_pixmap = new QPixmap(x);
}

void Impix::setImage(const QImage &x)
{
	if(v_image)
		delete v_image;
	v_image = new QImage(x);
}

bool Impix::load(const QString &name, const QStringList &dirs)
{
	unload();

	QImage i;
	QPixmap p;
	for(QStringList::ConstIterator it = dirs.begin(); it != dirs.end(); ++it) {
		QFileInfo fi(*it);
		if(fi.isDir()) {
			QString str = (*it) + "/" + name + ".png";
			if(i.load(str) == true) {
				// this is weird, but it gets the mask right on Windows
				//p.convertFromImage(i.copy(0,0,i.width(),i.height()));
				if(!p.convertFromImage(i))
					return false;

				setImage(i);
				setPixmap(p);
				return true;
			}
		}
		else if(fi.extension(false) == "zip") {
			UnZip z(*it);
			if(!z.open())
				continue;

			QByteArray a;
			if(!z.readFile(name + ".png", &a, 1024*128))
				continue;

			if(i.loadFromData(a) == true) {
				// this is weird, but it gets the mask right on Windows
				//p.convertFromImage(i.copy(0,0,i.width(),i.height()));
				if(!p.convertFromImage(i))
					return false;

				setImage(i);
				setPixmap(p);
				return true;
			}
		}
	}

        return false;
}

void Impix::unload()
{
	if(!v_pixmap)
		return;

	delete v_pixmap;
	v_pixmap = 0;
	delete v_image;
	v_image = 0;
}


/****************************************************************************
  Anim - handles animations
****************************************************************************/
Anim::Anim()
{
	seq.setAutoDelete(true);
}

Anim::~Anim()
{
}

bool Anim::load(const QString &name, const QStringList &dirs)
{
	unload();

	QImage image;
	QStringList::ConstIterator it;
	for(it = dirs.begin(); it != dirs.end(); ++it) {
		QFileInfo fi(*it);
		if(fi.isDir()) {
			QString str = (*it) + "/" + name + ".png";
			if(image.load(str))
				break;
		}
		else if(fi.extension(false) == "zip") {
			UnZip z(*it);
			if(!z.open())
				continue;

			QByteArray a;
			if(!z.readFile(name + ".png", &a, 1024*128))
				continue;

			if(image.loadFromData(a))
				break;
		}
	}
	if(it == dirs.end())
		return false;

	int n;
	for(n = 0; ((n+1) * 16) <= image.width(); ++n) {
		seq.resize(n+1);
		seq.insert(n, new Impix(image.copy(n * 16, 0, 16, 16)) );
	}

	// hopefully this never happens
	if(n == 0) {
		QPixmap p(16,16);
		p.fill(Qt::black);

		seq.resize(1);
		seq.insert(0, new Impix(p));

		v_isAnim = false;
		v_numFrames = 0;
		return false;
	}

	v_isAnim = false;
	v_numFrames = n - 1;
	if(v_numFrames > 0)
		v_isAnim = true;
		
	return true;
}

void Anim::unload()
{
	seq.clear();
	seq.resize(0);
}

Impix & Anim::frame(int x) const
{
	// if we don't have the frame, just return the base
	if(!seq[x+1])
		return *seq[0];

	return *seq[x+1];
}

Impix & Anim::base() const
{
	return *seq[0];
}


/****************************************************************************
  Emoticon
****************************************************************************/
Emoticon::Emoticon(const QString &name, const QRegExp &r)
:Impix()
{
	v_name = name;
	v_reg = r;
}

const QString & Emoticon::name() const
{
	return v_name;
}

const QRegExp & Emoticon::regExp() const
{
	return v_reg;
}

void Emoticon::setName(const QString &s)
{
	v_name = s;
}

void Emoticon::setRegExp(const QRegExp &r)
{
	v_reg = r;
}


/****************************************************************************
  EmoticonList
****************************************************************************/
EmoticonList::EmoticonList()
:QValueList<Emoticon>()
{
}

bool EmoticonList::add(const QStringList &dirs, const QString &name, const QRegExp &reg)
{
	Emoticon e(name, reg);
	bool ok;
	ok = e.load(QString("emoticon_") + name, dirs);
	append(e);
	return ok;
}

EmoticonList::Iterator EmoticonList::find(const QString &s)
{
	Iterator it;

	for(it = begin(); it != end(); ++it) {
		if((*it).name() == s)
			break;
	}
	return it;
}

QMimeSourceFactory *EmoticonList::generateFactory() const
{
	QMimeSourceFactory *m = new QMimeSourceFactory;
	addToFactory(m);
	return m;
}

void EmoticonList::addToFactory(QMimeSourceFactory *m) const
{
	for(ConstIterator it = begin(); it != end(); ++it) {
		const Emoticon &e = *it;
		if(!e.isNull())
			m->setImage(e.name(), e);
	}
}


/****************************************************************************
  PsiIconSet - icon set class
****************************************************************************/
PsiIconSet::PsiIconSet()
{
}

const Impix & PsiIconSet::status(const Status &s) const
{
	if(!s.isAvailable())
		return offline;
	else if(s.isInvisible())
		return invisible;
	else {
		if(s.show() == "away")
			return away;
		else if(s.show() == "xa")
			return xa;
		else if(s.show() == "dnd")
			return dnd;
		else
			return online;
	}
}

const Impix & PsiIconSet::status(int x) const
{
	switch(x) {
		case STATUS_OFFLINE:   return offline;
		case STATUS_AWAY:      return away;
		case STATUS_XA:        return xa;
		case STATUS_DND:       return dnd;
		case STATUS_INVISIBLE: return invisible;

		case STATUS_ONLINE:
		default:               return online;
	}
}

Anim & PsiIconSet::event2anim(PsiEvent *e)
{
	if(e->type() == PsiEvent::Message) {
		MessageEvent *me = (MessageEvent *)e;
		const Message &m = me->message();
		if(m.type() == "headline")
			return headline;
		else if(m.type() == "chat")
			return chat;
		else if(m.type() == "error")
			return system;
		else
			return message;
	}
	else if(e->type() == PsiEvent::PGP) {
		return needkey;
	}
	else {
		return system;
	}
}

bool PsiIconSet::load(const QString &name)
{
	if(!loadSet(name)) {
		unload();
		return false;
	}

	return true;
}

bool PsiIconSet::loadSet(const QString &name)
{
	QStringList dirs;
	QString subdir = "/iconsets/" + name;

	dirs += "." + subdir;
	dirs += g.pathHome + subdir;
	dirs += g.pathBase + subdir;
	dirs += g.pathBase + "/iconsets/stellar";

	// load animations
	if(!message.load("message", dirs)) return false;
	if(!system.load("system", dirs)) return false;
	if(!chat.load("chat", dirs)) return false;
	if(!headline.load("headline", dirs)) return false;
	if(!conn.load("connect", dirs)) return false;
	if(!needkey.load("pgp", dirs)) return false;

	// load pixmaps
	if(!online.load("online", dirs)) return false;
	if(!offline.load("offline", dirs)) return false;
	if(!away.load("away", dirs)) return false;
	if(!xa.load("xa", dirs)) return false;
	if(!dnd.load("dnd", dirs)) return false;
	if(!invisible.load("invisible", dirs)) return false;
	if(!noauth.load("noauth", dirs)) return false;
	if(!perr.load("perr", dirs)) return false;
	if(!arrow[0].load("groupclose", dirs)) return false;
	if(!arrow[1].load("groupopen", dirs)) return false;
	if(!arrow[2].load("groupempty", dirs)) return false;
	if(!add.load("add", dirs)) return false;
	if(!remove.load("remove", dirs)) return false;
	if(!send.load("send", dirs)) return false;
	if(!history.load("history", dirs)) return false;
	if(!info.load("info", dirs)) return false;
	if(!profile.load("changeacc", dirs)) return false;
	if(!url.load("url", dirs)) return false;
	if(!groupchat.load("groupchat", dirs)) return false;
	if(!ask.load("ask", dirs)) return false;

	if(!psimain.load("psimain", dirs)) return false;
	if(!ssl_yes.load("ssl_yes", dirs)) return false;
	if(!ssl_no.load("ssl_no", dirs)) return false;
	if(!chatclear.load("chatclear", dirs)) return false;
	if(!account.load("account", dirs)) return false;

	if(!options.load("options", dirs)) return false;
	if(!play_sounds.load("play_sounds", dirs)) return false;
	if(!quit.load("quit", dirs)) return false;
	if(!help.load("help", dirs)) return false;
	if(!xml.load("xml", dirs)) return false;
	if(!arrow_left.load("arrow_left", dirs)) return false;
	if(!arrow_right.load("arrow_right", dirs)) return false;
	if(!arrow_up.load("arrow_up", dirs)) return false;
	if(!arrow_down.load("arrow_down", dirs)) return false;

	if(!browse.load("jabber", dirs)) return false;
	if(!registerAgent.load("register", dirs)) return false;
	if(!search.load("search", dirs)) return false;
	if(!getTime.load("time", dirs)) return false;
	if(!getLastTime.load("time", dirs)) return false;
	if(!refresh.load("reload", dirs)) return false;
	if(!stop.load("stop", dirs)) return false;
	if(!getVersion.load("jabber", dirs)) return false;

	// load emoticons that are available
	emoticons.add(dirs, "smile",       QRegExp(":\\)|:-\\)"));
	emoticons.add(dirs, "wink",        QRegExp(";\\)|;-\\)"));
	emoticons.add(dirs, "tongue",      QRegExp(":p|:P|:-p|:-P"));
	emoticons.add(dirs, "biggrin",     QRegExp(":D|:-D|:>|:->|:d|:-d"));
	emoticons.add(dirs, "unhappy",     QRegExp(":\\(|:-\\("));
	emoticons.add(dirs, "cry",         QRegExp(":'-\\(|:'\\(|;-\\(|;\\("));
	emoticons.add(dirs, "oh",          QRegExp(":o|:O|:-o|:-O"));
	emoticons.add(dirs, "angry",       QRegExp(":-@|:@"));
	emoticons.add(dirs, "blush",       QRegExp(":\\$|:-\\$"));
	emoticons.add(dirs, "heart",       QRegExp("\\(l\\)|\\(L\\)"));
	emoticons.add(dirs, "brheart",     QRegExp("\\(u\\)|\\(U\\)"));
	emoticons.add(dirs, "yes",         QRegExp("\\(y\\)|\\(Y\\)"));
	emoticons.add(dirs, "no",          QRegExp("\\(n\\)|\\(N\\)"));
	emoticons.add(dirs, "boy",         QRegExp("\\(z\\)|\\(Z\\)"));
	emoticons.add(dirs, "girl",        QRegExp("\\(x\\)|\\(X\\)"));
	emoticons.add(dirs, "pussy",       QRegExp("\\(@\\)"));
	emoticons.add(dirs, "hugleft",     QRegExp("\\(\\}\\)"));
	emoticons.add(dirs, "hugright",    QRegExp("\\(\\{\\)"));
	emoticons.add(dirs, "devil",       QRegExp("\\(6\\)"));
	emoticons.add(dirs, "rainbow",     QRegExp("\\(R\\)|\\(r\\)"));
	emoticons.add(dirs, "bat",         QRegExp(":\\[|:-\\["));
	emoticons.add(dirs, "brflower",    QRegExp("\\(w\\)|\\(W\\)"));
	emoticons.add(dirs, "flower",      QRegExp("\\(f\\)|\\(F\\)"));
	emoticons.add(dirs, "photo",       QRegExp("\\(p\\)|\\(P\\)"));
	emoticons.add(dirs, "phone",       QRegExp("\\(t\\)|\\(T\\)"));
	emoticons.add(dirs, "star",        QRegExp("\\(\\*\\)"));
	emoticons.add(dirs, "music",       QRegExp("\\(8\\)"));
	emoticons.add(dirs, "lamp",        QRegExp("\\(i\\)|\\(I\\)"));
	emoticons.add(dirs, "beer",        QRegExp("\\(b\\)|\\(B\\)"));
	emoticons.add(dirs, "drink",       QRegExp("\\(d\\)|\\(D\\)"));
	emoticons.add(dirs, "coffee",      QRegExp("\\(c\\)|\\(C\\)"));
	emoticons.add(dirs, "cuffs",       QRegExp("\\(\\%\\)"));
	emoticons.add(dirs, "mail",        QRegExp("\\(e\\)|\\(E\\)"));
	emoticons.add(dirs, "stare",       QRegExp(":\\||:-\\|"));
	emoticons.add(dirs, "frowning",    QRegExp(":s|:-s|:S|:-S"));
	emoticons.add(dirs, "coolglasses", QRegExp("\\(h\\)|\\(H\\)"));
	emoticons.add(dirs, "kiss",        QRegExp("\\(k\\)|\\(K\\)"));

	QMimeSourceFactory *m = QMimeSourceFactory::defaultFactory();
	m->setPixmap("online", online);
	m->setPixmap("offline", offline);
	m->setPixmap("away", away);
	m->setPixmap("xa", xa);
	m->setPixmap("dnd", dnd);
	m->setPixmap("invisible", invisible);

	return true;
}

void PsiIconSet::unload()
{
	message.unload();
	system.unload();
	chat.unload();
	headline.unload();
	conn.unload();
	needkey.unload();

	online.unload();
	offline.unload();
	away.unload();
	xa.unload();
	dnd.unload();
	invisible.unload();
	noauth.unload();
	perr.unload();
	arrow[0].unload();
	arrow[1].unload();
	arrow[2].unload();
	add.unload();
	remove.unload();
	send.unload();
	history.unload();
	info.unload();
	profile.unload();
	url.unload();
	groupchat.unload();
	ask.unload();

	psimain.unload();
	ssl_yes.unload();
	ssl_no.unload();
	chatclear.unload();
	account.unload();

	emoticons.clear();
}

void PsiIconSet::getIconSetsList(QStringList &names, QStringList &descriptions)
{
	QStringList dirs;
	QString subdir = "/iconsets";

	dirs += "." + subdir;
	dirs += g.pathHome + subdir;
	dirs += g.pathBase + subdir;

	for(QStringList::Iterator it = dirs.begin(); it != dirs.end(); ++it) {
		if(!QFile::exists(*it))
			continue;
		QDir d(*it);
		QStringList entries = d.entryList();
		for(QStringList::Iterator it2 = entries.begin(); it2 != entries.end(); ++it2) {
			// make sure we don't have it
			bool found = FALSE;
			for(QStringList::Iterator it3 = names.begin(); it3 != names.end(); ++it3) {
				if(*it3 == *it2) {
					found = TRUE;
					break;
				}
			}
			if(found)
				continue;

			QString fname = *it + '/' + *it2;

			QFileInfo fi(fname);
			QString desc;
			if(fi.isDir()) {
				QFile f(fname + "/desc");
				if(!f.open(IO_ReadOnly))
					continue;
				QCString line;
				line.resize(256);
				int n = f.readLine(line.data(), 255);
				if(n == -1)
					continue;
				line[n] = 0;
				if(line.at(line.length()-1) == '\n')
					line.at(line.length()-1) = 0;

				desc = QString::fromUtf8(line);
			}
			else if(fi.extension(false) == "zip") {
				UnZip z(fname);
				if(!z.open())
					continue;
				QByteArray a;
				if(!z.readFile("desc", &a, 256))
					continue;

				QCString cstr(a.size()+1);
				memcpy(cstr.data(), a.data(), a.size());
				QString str = QString::fromUtf8(cstr);

				QStringList lines = QStringList::split("\n", str);
				desc = lines[0] + " {ZIP}";
			}
			else
				continue;

			names += *it2;
			descriptions += desc;
		}
	}
}


/****************************************************************************
  AnimatedButtom - an animated button!
****************************************************************************/
AnimatedButton::AnimatedButton(QWidget *parent, const char *name, bool _animOnly)
:QPushButton(parent, name)
{
	setWFlags(getWFlags() | WRepaintNoErase);

	animStep = 0;
	v_isActive = FALSE;
	animOnly = _animOnly;
}

void AnimatedButton::setAnim(Anim *_anim)
{
	int alertStyle = option.alertStyle;
	if(animOnly)
		alertStyle = 2;

	if(_anim) {
		anim = _anim;
		animStep = 0;
		if(alertStyle == 0)
			changeIcon(anim->base());
		if(!v_isActive) {
			v_isActive = TRUE;

			t = new QTimer(this);
			connect(t, SIGNAL(timeout()), SLOT(animate()));
			t->start(120);
		}
	}
	else {
		// no icon
	        QIconSet icon;
		setIconSet(icon);

		anim = 0;
		if(v_isActive) {
			v_isActive = FALSE;
			delete t;
		}
	}
}

void AnimatedButton::animate()
{
	int alertStyle = option.alertStyle;
	if(animOnly)
		alertStyle = 2;

	if(anim->isAnim() && alertStyle == 2) {
		changeIcon(anim->frame(animStep));

		++animStep;
		if(animStep > anim->numFrames()-1)
			animStep = 0;
	}
	else if(alertStyle != 0) {
		if(animStep == 0)
			changeIcon(anim->base());
		else if(animStep == 4)
			changeIcon(*pix_blank);

		++animStep;
		if(animStep == 8)
			animStep = 0;
	}
}

void AnimatedButton::changeIcon(const QPixmap &pix)
{
	//setUpdatesEnabled(FALSE);
	setIconSet(pix);
	//setUpdatesEnabled(TRUE);
	//update();
}
