/****************************************************************************
** mainwin_p.cpp - classes used privately by the main window.
** Copyright (C) 2001, 2002  Justin Karneges, Michail Pishchagin
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
**
****************************************************************************/

#include "mainwin_p.h"

#include <qstyle.h>
#include <qtoolbar.h>
#include <qtimer.h>
#include <qsignalmapper.h>
#include "psiaccount.h"

//*******************************************************
//  MPushButton
//*******************************************************
MPushButton::MPushButton(QWidget *parent, const char *name)
:AnimatedButton(parent, name, true)
{
	//setMinimumWidth(48);
	setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));

	hasTip = FALSE;
}

void MPushButton::setText(const QString &_text)
{
	if(hasTip) {
		QToolTip::remove(this);
		hasTip = FALSE;
	}

	QPushButton::setText(_text);
}

void MPushButton::drawButtonLabel(QPainter *p)
{
	// crazy code ahead!  watch out for potholes and deer.

	// this gets us the width of the "text area" on the button.
	// adapted from qt/src/styles/qcommonstyle.cpp and qt/src/widgets/qpushbutton.cpp
	QRect r = style().subRect(QStyle::SR_PushButtonContents, this);
	if(isMenuButton())
		r.setWidth(r.width() - style().pixelMetric(QStyle::PM_MenuButtonIndicator, this));
	if(iconSet() && !iconSet()->isNull())
		r.setWidth(r.width() - (iconSet()->pixmap(QIconSet::Small, QIconSet::Normal, QIconSet::Off).width() + 4));

	// font metrics
	QFontMetrics fm(font());

	// w1 = width of button text, w2 = width of text area
	int w1 = fm.width(text());
	int w2 = r.width();

	// backup original text
	QString oldtext = text();

	// button text larger than what will fit?
	if(w1 > w2) {
		if(!hasTip) {
			QToolTip::add(this, text());
			hasTip = TRUE;
		}

		// make a string that fits
		bool found = FALSE;
		QString newtext;
		int n;
		for(n = oldtext.length(); n > 0; --n) {
			if(fm.width(oldtext, n) < w2) {
				found = TRUE;
				break;
			}
		}
		if(found)
			newtext = oldtext.mid(0, n);
		else
			newtext = "";

		// set the new text that fits.  updates must be off, or we recurse.
		setUpdatesEnabled(FALSE);
		QButton::setText(newtext);
		setUpdatesEnabled(TRUE);
	}
	else {
		if(hasTip) {
			QToolTip::remove(this);
			hasTip = FALSE;
		}
	}

	// draw!
	QPushButton::drawButtonLabel(p);

	// restore original button text now that we are done drawing.
	setUpdatesEnabled(FALSE);
	QButton::setText(oldtext);
	setUpdatesEnabled(TRUE);
}

void MPushButton::paintEvent(QPaintEvent *e)
{
	QPushButton::paintEvent(e);
}

QSize MPushButton::minimumSizeHint() const
{
	return QPushButton::minimumSizeHint();
}

QSize MPushButton::sizeHint() const
{
	return QPushButton::sizeHint();
}

//*******************************************************
//  PopupAction
//*******************************************************
class PopupAction::Private : public QObject
{
public:
	QPopupMenu *menu;
	QSizePolicy size;
	QPtrList<MPushButton> buttons;
	Anim *anim;
	bool usePixmap;
	bool minimumSize;

	Private (QPopupMenu *_menu, QObject *parent)
	: QObject (parent)
	{
		menu = _menu;
		anim = 0;
		usePixmap = false;
		minimumSize = false;
	}
};

PopupAction::PopupAction (const QString &label, QPopupMenu *_menu, QObject *parent, const char *name)
: QAction (label, label, 0, parent, name)
{
	d = new Private (_menu, this);
}

void PopupAction::setSizePolicy (const QSizePolicy &p)
{
	d->size = p;
}

void PopupAction::setAnim (Anim *a)
{
	d->anim = a;
	for ( QPtrListIterator<MPushButton> it(d->buttons); it.current(); ++it ) {
		MPushButton *btn = it.current();
		btn->setAnim (a);
	}
}

void PopupAction::setIconSet (const QIconSet &icon)
{
	QAction::setIconSet (icon);
	for ( QPtrListIterator<MPushButton> it(d->buttons); it.current(); ++it ) {
		MPushButton *btn = it.current();
		btn->setIconSet (icon);
	}
}

void PopupAction::setPixmap (const QPixmap &pix)
{
	d->usePixmap = true;
	QAction::setIconSet ( pix );
	for ( QPtrListIterator<MPushButton> it(d->buttons); it.current(); ++it ) {
		MPushButton *btn = it.current();
		btn->setPixmap (pix);
	}
}

void PopupAction::setText (const QString &text)
{
	QAction::setText (text);
	for ( QPtrListIterator<MPushButton> it(d->buttons); it.current(); ++it ) {
		MPushButton *btn = it.current();
		btn->setText (text);
	}
}

bool PopupAction::addTo (QWidget *w)
{
	if ( w->inherits("QPopupMenu") ) {
		QPopupMenu *m = (QPopupMenu*)w;
		m->insertItem (iconSet(), menuText(), d->menu);
	}
	else if ( w->inherits("QToolBar") ) {
		QCString bname = name() + QCString("_action_button");
		MPushButton *btn = new MPushButton ( (QToolBar*)w, bname );
		d->buttons.append ( btn );
		btn->setPopup ( d->menu );
		btn->setAnim ( d->anim );
		btn->setText ( text() );
		if ( d->usePixmap )
			btn->setPixmap ( iconSet().pixmap() );
		else
			btn->setIconSet ( iconSet() );
		btn->setSizePolicy ( d->size );
		btn->setEnabled ( isEnabled() );

		connect( btn, SIGNAL( destroyed() ), SLOT( objectDestroyed() ) );
	}
	else
		return false;
	return true;
}

void PopupAction::objectDestroyed ()
{
	const QObject *obj = sender();
	d->buttons.removeRef( (MPushButton*) obj );
}

void PopupAction::setEnabled (bool e)
{
	QAction::setEnabled (e);
	for ( QPtrListIterator<MPushButton> it(d->buttons); it.current(); ++it ) {
		MPushButton *btn = it.current();
		btn->setEnabled (e);
	}
}

//*******************************************************
//  MLabel
//*******************************************************
MLabel::MLabel(QWidget *parent, const char *name)
:QLabel(parent, name)
{
}

void MLabel::mouseReleaseEvent(QMouseEvent *e)
{
	clicked(e->button());
	e->ignore();
}

void MLabel::mouseDoubleClickEvent(QMouseEvent *e)
{
	if(e->button() == LeftButton)
		doubleClicked();

	e->ignore();
}

//----------------------------------------------------------------------------
// MTray
//----------------------------------------------------------------------------
MTray::MTray(const QImage &icon, const QString &tip, QPopupMenu *popup, QObject *parent)
:QObject(parent)
{
	anim = 0;

	trayTimer = new QTimer(this);
	connect(trayTimer, SIGNAL(timeout()), SLOT(animate()));

	ti = new TrayIcon(makeIcon(icon), tip, popup);
	connect(ti, SIGNAL(clicked(const QPoint &, int)), SIGNAL(clicked(const QPoint &, int)));
	connect(ti, SIGNAL(doubleClicked(const QPoint &)), SIGNAL(doubleClicked(const QPoint &)));
	connect(ti, SIGNAL(closed()), SIGNAL(closed()));
	ti->show();
}

MTray::~MTray()
{
	delete ti;
}

void MTray::setToolTip(const QString &str)
{
	ti->setToolTip(str);
}

void MTray::setImage(const QImage &i)
{
	trayTimer->stop();
	anim = 0;
	ti->setIcon(makeIcon(i));
}

void MTray::setAnim(Anim *a)
{
	anim = a;
	animStep = 0;
	if(option.alertStyle == 0)
		ti->setIcon(makeIcon(anim->base()));

	if(!trayTimer->isActive())
		trayTimer->start(120);
}

bool MTray::isAnimating() const
{
	return (anim ? true: false);
}

void MTray::show()
{
	ti->show();
}

void MTray::hide()
{
	ti->hide();
}

QPixmap MTray::makeIcon(const QImage &_in)
{
#ifdef Q_WS_X11
	/*if(1) {
		QImage real = wmdock[0];
		real.detach();
		QImage in = _in;
		in.detach();

		// make sure it is no bigger than 16x16
		if(in.width() > 16 || in.height() > 16)
			in = in.smoothScale(16,16);

		int xo = 40;
		int yo = 40;

		int n, n2;

		// draw a border
		for(n = 0; n < 22; ++n) {
			real.setPixel(n + xo, yo, qRgb(0x80,0x80,0x80));
			real.setPixel(n + xo, yo + 21, qRgb(0x00,0x00,0x00));
			real.setPixel(xo, yo + n, qRgb(0x80,0x80,0x80));
			real.setPixel(xo + 21, yo + n, qRgb(0x00,0x00,0x00));
		}

		xo += 3;
		yo += 3;

		// draw a dropshadow
		for(n2 = 0; n2 < in.height(); ++n2) {
			for(n = 0; n < in.width(); ++n) {
				if(qAlpha(_in.pixel(n,n2))) {
					real.setPixel(n + xo + 2, n2 + yo + 2, qRgb(0x80,0x80,0x80));
				}
			}
		}
		for(n2 = 0; n2 < in.height(); ++n2) {
			for(n = 0; n < in.width(); ++n) {
				if(qAlpha(_in.pixel(n,n2))) {
					real.setPixel(n + xo, n2 + yo, _in.pixel(n,n2));
				}
			}
		}

		QPixmap icon;
		icon.convertFromImage(real);
		return icon;
	}
	else {*/
		// on X11, the KDE dock is 22x22.  let's make our icon "seem" bigger.
		QImage real(22,22,32);
		//QImage in = _in.convertToImage();
		QImage in = _in;
		in.detach();
		real.fill(0);
		real.setAlphaBuffer(true);

		// make sure it is no bigger than 16x16
		if(in.width() > 16 || in.height() > 16)
			in = in.smoothScale(16,16);

		int xo = (real.width() - in.width()) / 2;
		int yo = (real.height() - in.height()) / 2;

		int n, n2;

		// draw a border
		/*for(n = 0; n < 22; ++n) {
			real.setPixel(n,  0, qRgb(0x80,0x80,0x80));
			real.setPixel(n, 21, qRgb(0x00,0x00,0x00));
			real.setPixel( 0, n, qRgb(0x80,0x80,0x80));
			real.setPixel(21, n, qRgb(0x00,0x00,0x00));
		}*/

		// draw a dropshadow
		for(n2 = 0; n2 < in.height(); ++n2) {
			for(n = 0; n < in.width(); ++n) {
				if(qAlpha(in.pixel(n,n2))) {
					int x = n + xo + 2;
					int y = n2 + yo + 2;
					real.setPixel(x, y, qRgba(0x80,0x80,0x80,0xff));
				}
			}
		}
		// draw the image
		for(n2 = 0; n2 < in.height(); ++n2) {
			for(n = 0; n < in.width(); ++n) {
				if(qAlpha(in.pixel(n,n2))) {
					QRgb c = in.pixel(n, n2);
					real.setPixel(n + xo, n2 + yo, qRgba(qRed(c), qGreen(c), qBlue(c), 0xff));
				}
			}
		}
		// create the alpha layer
		for(n2 = real.height()-2; n2 >= 0; --n2) {
			for(n = 0; n < real.width(); ++n) {
				uint c = real.pixel(n, n2);
				if(c > 0) {
					QRgb old = real.pixel(n, n2);
					real.setPixel(n, n2, qRgba(qRed(old), qGreen(old), qBlue(old), 0xff));
				}
			}
		}

		QPixmap icon;
		icon.convertFromImage(real);
		return icon;
	//}
#else
	QPixmap icon;
	icon.convertFromImage(_in);
	return icon;
#endif
}

void MTray::animate()
{
	if(anim->isAnim() && option.alertStyle == 2) {
		ti->setIcon(makeIcon(anim->frame(animStep)));

		++animStep;
		if(animStep > anim->numFrames()-1)
			animStep = 0;
	}
	else if(option.alertStyle != 0) {
		if(animStep == 0)
			ti->setIcon(makeIcon(anim->base()));
		else if(animStep == 4) {
			QImage blank(16,16,32);
			blank.fill(0);
			blank.setAlphaBuffer(true);
			ti->setIcon(makeIcon(blank));
		}

		++animStep;
		if(animStep == 8)
			animStep = 0;
	}
}

//----------------------------------------------------------------------------
// MAction
//----------------------------------------------------------------------------
class MAction::Private : public QObject
{
public:
	int id;
	PsiCon *psi;

	QPtrList<QToolButton> buttons;
	QSignalMapper *sm;

	Private (int _id, PsiCon *_psi, QObject *parent)
	: QObject (parent)
	{
		id = _id;
		psi = _psi;
		sm = new QSignalMapper(this, "MAction::Private::SignalMapper");
	}

	QPopupMenu *subMenu(QWidget *p)
	{
		QPopupMenu *sm = new QPopupMenu (p);
		uint i = 0;
		for ( PsiAccountListIt it(psi->accountList()); it.current(); ++it, i++ )
		{
			PsiAccount *acc = it.current();
			sm->insertItem( acc->name(), parent(), SLOT(itemActivated(int)), 0, id*1000 + i );
			sm->setItemParameter ( id*1000 + i, i );
		}
		return sm;
	}
};

MAction::MAction(const QIconSet &i, const QString &s, int id, PsiCon *psi, QObject *parent)
:QAction(s, s, 0, parent)
{
	init (i, id, psi);
}

MAction::MAction(const QString &s, int id, PsiCon *psi, QObject *parent)
:QAction(s, s, 0, parent)
{
	init (QIconSet(), id, psi);
}

void MAction::init(const QIconSet &i, int id, PsiCon *psi)
{
	d = new Private(id, psi, this);
	setIconSet (i);
	connect(psi, SIGNAL(accountCountChanged()), SLOT(numAccountsChanged()));
	setEnabled ( !d->psi->accountList().isEmpty() );
	connect (d->sm, SIGNAL(mapped(int)), SLOT(itemActivated(int)));
}

bool MAction::addTo(QWidget *w)
{
	if ( w->inherits("QPopupMenu") )
	{
		QPopupMenu *menu = (QPopupMenu*)w;
		if ( d->psi->accountList().count() < 2 ) {
			menu->insertItem ( iconSet(), menuText(), this, SLOT(itemActivated(int)), 0, d->id*1000 + 0 );
			menu->setItemEnabled (d->id*1000 + 0, isEnabled());
			menu->setItemParameter ( d->id*1000 + 0, 0 );
		}
		else 
			menu->insertItem(iconSet(), menuText(), d->subMenu(w));
	}
	else if ( w->inherits("QToolBar") )
	{
		QToolButton *btn = new QToolButton ( (QToolBar*)w, "MAction_button" );
		d->buttons.append ( btn );
		if ( d->psi->accountList().count() >= 2 )
			btn->setPopup ( d->subMenu(btn) );
		else
			btn->setPopup (0);
		
		connect (btn, SIGNAL(clicked()), d->sm, SLOT(map()));
		d->sm->setMapping (btn, 0);
		
		btn->setEnabled ( isEnabled() );
		btn->setIconSet ( iconSet() );
		btn->setPopupDelay (1); // the popup will be displayed immediately

		QString tt, str = menuText();
		for (int i = 0; i < (int)str.length(); i++)
			if ( str[i] == '&' && str[i+1] != '&' )
				continue;
			else
				tt += str[i];

		QToolTip::add (btn, tt);

		connect( btn, SIGNAL( destroyed() ), SLOT( objectDestroyed() ) );
	}
	else
		return false;
	return true;
}

void MAction::itemActivated(int n)
{
	PsiAccountList list = (PsiAccountList)d->psi->accountList();
	
	if (n >= (int)list.count()) // just in case
		return;

	emit activated((PsiAccount *)list.at(n), d->id);
}

void MAction::numAccountsChanged()
{
	setEnabled ( !d->psi->accountList().isEmpty() );

	for ( QPtrListIterator<QToolButton> it(d->buttons); it.current(); ++it ) {
		QToolButton *btn = it.current();

		if ( btn->popup() )
			delete btn->popup();
		btn->setPopup (0);

		if ( d->psi->accountList().count() >= 2 )
			btn->setPopup ( d->subMenu(btn) );
	}
}

void MAction::objectDestroyed()
{
	const QObject *obj = sender();
	d->buttons.removeRef( (QToolButton*) obj );
}

void MAction::setEnabled (bool e)
{
	QAction::setEnabled (e);
	for ( QPtrListIterator<QToolButton> it(d->buttons); it.current(); ++it ) {
		QToolButton *btn = it.current();

		btn->setEnabled (e);
	}
}

void MAction::setIconSet (const QIconSet &icon)
{
	QAction::setIconSet (icon);
	for ( QPtrListIterator<QToolButton> it(d->buttons); it.current(); ++it ) {
		QToolButton *btn = it.current();

		btn->setIconSet (icon);
	}
}
